﻿using System;
using System.Collections.Generic;
using TcHmiSrv;
using TcHmiSrv.DynamicSymbols;

namespace ListSymbolSampleExtension
{
    public class ListSymbolSampleExtension
    {
        // Managed instance of ITcHmiSrvExtHost
        private ITcHmiSrvExtHost _host = null;

        private object _shutdownLock = new object();
        private bool _isShuttingDown = false;

        // Listeners to handle events
        private TcHmiSrvRequestListener _requestListener = new TcHmiSrvRequestListener();
        private TcHmiSrvShutdownListener _shutdownListener = new TcHmiSrvShutdownListener();

        private Log _logger = null;
        private Data _data = new Data();

        private List<Symbol> symbols = new List<Symbol>();

        // Initialize extension
        public ErrorValue Init(ITcHmiSrvExtHost extHost, Context ctx)
        {
            _logger = new Log(extHost);

            try
            {
                _host = extHost;

                // INSERT INIT CODE HERE
                // Define custom variables
                _data.StringArray.Add("one");
                _data.StringArray.Add("two");
                _data.StringArray.Add("three");

                Struct structSample = new Struct();
                structSample._string_value = "stringmember";
                structSample._int_value = 42;
                _data.StructSample = structSample;

                // Add symbols
                symbols.Add(new Symbol("customArrayValue", typeof(List<string>)));
                symbols.Add(new Symbol("customStructValue", typeof(Struct)));

                // Add event handlers
                _requestListener.OnRequest += OnRequest;
                _shutdownListener.OnShutdown += OnShutdown;

                // Register listeners
                _host.register_listener(ctx, _requestListener);
                _host.register_listener(ctx, _shutdownListener);

                _logger.Send(ctx, "MESSAGE_INIT");
                return ErrorValue.HMI_SUCCESS;
            }
            catch (Exception ex)
            {
                _logger.Send(ctx, "ERROR_INIT", ex.Message, Severity.Severity_Error);
                return ErrorValue.HMI_E_EXTENSION_LOAD;
            }
        }

        private ErrorValue OnRequest(object sender, TcHmiSrvRequestListener.OnRequestEventArgs e)
        {
            ErrorValue ret = ErrorValue.HMI_SUCCESS;
            Context ctx = e.Context;
            CommandGroup commands = e.Commands;

            try
            {
                commands.Result = ExtensionErrorValue.HMI_EXT_SUCCESS;
                string mapping = "";

                foreach (Command command in commands) // Iterate the commands
                {
                    mapping = command.Mapping; // Get the name of the current command

                    try
                    {
                        if (mapping.Equals("customArrayValue")) // Check if the customArrayValue symbol is requested
                        {
                            command.ReadValue = TcHmiSerializer.SerializeObject(TcHmiUniversalResolver.ResolveObject(_data.StringArray, command.Path));
                        }
                        else if (mapping.Equals("customStructValue")) // Check if the customStructValue symbol is requested
                        {
                            command.ReadValue = TcHmiSerializer.SerializeObject(TcHmiUniversalResolver.ResolveObject(_data.StructSample, command.Path));
                        }
                        else if (mapping.Equals("GetSchema"))
                        {
                            GetSchema(command);
                        }
                        else if (mapping.Equals("ListSymbols"))
                        {
                            ListSymbols(command);
                        }
                    }
                    catch (Exception ex)
                    {
                        command.ExtensionResult = ExtensionErrorValue.HMI_EXT_FAIL;
                        command.ResultString = _logger.Localize(e.Context, "ERROR_CALL_COMMAND", new string[] { mapping, ex.Message });
                    }
                    finally
                    {
                        if (commands.Result == ExtensionErrorValue.HMI_EXT_SUCCESS && command.ExtensionResult != ExtensionErrorValue.HMI_EXT_SUCCESS)
                        {
                            commands.Result = command.ExtensionResult;
                        }
                    }
                }
            }
            catch
            {
                commands.Result = ExtensionErrorValue.HMI_EXT_FAIL;
            }
            finally
            {
                if (commands.Result != ExtensionErrorValue.HMI_EXT_SUCCESS)
                {
                    // Reset the read value of the commands to prevent the server from sending invalid data
                    foreach (Command command in commands)
                    {
                        command.ReadValue = null;
                    }

                    ret = ErrorValue.HMI_E_EXTENSION;
                }
            }

            return ret;
        }

        // Shutdown extension and unregister all registered listeners
        private ErrorValue OnShutdown(object sender, TcHmiSrvShutdownListener.OnShutdownEventArgs e)
        {
            // If the extension does not shutdown after 10 seconds (blocking threads) OnShutdown will be called again
            lock (_shutdownLock)
            {
                if (_isShuttingDown)
                {
                    return ErrorValue.HMI_SUCCESS;
                }

                _isShuttingDown = true;

                Context ctx = e.Context;

                try
                {
                    // Unregister listeners
                    _host.unregister_listener(ctx, _requestListener);
                    _host.unregister_listener(ctx, _shutdownListener);

                    _logger.Send(ctx, "MESSAGE_SHUTDOWN");
                    return ErrorValue.HMI_SUCCESS;
                }
                catch (Exception ex)
                {
                    _logger.Send(ctx, "ERROR_SHUTDOWN", ex.Message, Severity.Severity_Error);
                    return ErrorValue.HMI_E_EXTENSION;
                }
            }
        }

        private ErrorValue GetSchema(Command command)
        {
            if (command.WriteValue.IsSet && command.WriteValue.Type == TcHmiSrv.ValueType.ValueType_String)
            {
                string path = path = command.WriteValue;

                foreach (Symbol symbol in symbols)
                {
                    if (path.StartsWith(symbol.Name))
                    {
                        path = RemoveFirstElement(path);

                        TcHmiSchemaGenerator.GetSchema(command,
                            symbol.ReadValue == null ? null : TcHmiUniversalResolver.ResolveType(symbol.ReadValue, path),
                            symbol.WriteValue, symbol.IsFunction,
                            symbol.Info.Members,
                            symbol.Info.BindingAttr,
                            symbol.Info.IgnoreElements,
                            symbol.Info.MaxDepth);

                        return ErrorValue.HMI_SUCCESS;
                    }
                }

                return ErrorValue.HMI_E_ENTRY_NOT_FOUND;
            }

            return ErrorValue.HMI_E_TYPE_MISMATCH;
        }

        private ErrorValue ListSymbols(Command command)
        {
            TcHmiSchemaGenerator.ListSymbols(command, symbols);
            return ErrorValue.HMI_SUCCESS;
        }

        private string RemoveFirstElement(string path)
        {
            Stack<string> parts = new Stack<string>(path.Split(new string[] { "::" }, StringSplitOptions.RemoveEmptyEntries));

            if (parts.Count == 0)
                throw new FormatException("The path is in an invalid format.");

            parts = new Stack<string>(parts);

            string first = parts.Pop();
            int begin = first.IndexOf('[');
            int end = first.IndexOf(']');

            if (begin > 0 && end > 0)
            {
                if (begin > end)
                    throw new FormatException("The path is in an invalid format.");

                parts.Push(first.Substring(begin, end - begin + 1));
            }

            return string.Join("::", parts);
        }
    }
}
